// transforme de Ingrid Daubechies  4 moments
// voir : http://www.bearcave.com/software/java/wavelets/daubechies/index.html

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>

using namespace std;

double  h0,  h1,  h2,  h3;
double  g0,  g1,  g2,  g3;
double Ih0, Ih1, Ih2, Ih3;
double Ig0, Ig1, Ig2, Ig3;

   /** Initialisation */
void DaubInit() {
    const double sqrt_3 = sqrt( 3 );
    const double denom = 4 * sqrt( 2 );
    //
    // forward transform scaling (smoothing) coefficients
    //
    h0 = (1 + sqrt_3)/denom;
    h1 = (3 + sqrt_3)/denom;
    h2 = (3 - sqrt_3)/denom;
    h3 = (1 - sqrt_3)/denom;
    //
    // forward transform wavelet coefficients
    //
    g0 =  h3;
    g1 = -h2;
    g2 =  h1;
    g3 = -h0;

    Ih0 = h2;
    Ih1 = g2;  // h1
    Ih2 = h0;
    Ih3 = g0;  // h3

    Ig0 = h3;
    Ig1 = g3;  // -h0
    Ig2 = h1;
    Ig3 = g1;  // -h2
}

/** Forward Daubechies D4 transform */
void Transform(double* a, const int n) {
    if (n >= 4) {
        int i, j;
        const int half = n >> 1;
        double* tmp = new double[n];
        for (i = 0, j = 0; j < n-3; j += 2, i++) {
            tmp[i]      = a[j]*h0 + a[j+1]*h1 + a[j+2]*h2 + a[j+3]*h3;
            tmp[i+half] = a[j]*g0 + a[j+1]*g1 + a[j+2]*g2 + a[j+3]*g3;
        }
        tmp[i]      = a[n-2]*h0 + a[n-1]*h1 + a[0]*h2 + a[1]*h3;
        tmp[i+half] = a[n-2]*g0 + a[n-1]*g1 + a[0]*g2 + a[1]*g3;
        for (i = 0; i < n; i++) a[i] = tmp[i];
        delete [] tmp;
    }
}

/** Inverse Daubechies D4 transform */
void InvTransform(double* a, const int n) {
    if (n >= 4) {
        int i, j;
        const int half = n >> 1;
        const int halfPls1 = half + 1;
        double* tmp = new double[n];
        //      last smooth val  last coef.  first smooth  first coef
        tmp[0] = a[half-1]*Ih0 + a[n-1]*Ih1 + a[0]*Ih2 + a[half]*Ih3;
        tmp[1] = a[half-1]*Ig0 + a[n-1]*Ig1 + a[0]*Ig2 + a[half]*Ig3;
        for (i = 0, j = 2; i < half-1; i++) {
	  //     smooth val     coef. val       smooth val    coef. val
	  tmp[j++] = a[i]*Ih0 + a[i+half]*Ih1 + a[i+1]*Ih2 + a[i+halfPls1]*Ih3;
	  tmp[j++] = a[i]*Ig0 + a[i+half]*Ig1 + a[i+1]*Ig2 + a[i+halfPls1]*Ig3;
        }
        for (i = 0; i < n; i++) a[i] = tmp[i];
        delete [] tmp;
    }
}

void DaubTrans(double* ts, int n) {
    int i;
    for (i = n; i >= 4; i >>= 1) {
        Transform(ts, i);
    }
}

void DaubInvTrans(double* coef, int n) {
    int i;
    for (i = 4; i <= n; i <<= 1) InvTransform(coef, i);
}

// Petit programme de dmonstration du dbruitage d'un signal par transforme en ondelettes

int main() {
    int i;
    const int n=8;
    double d, br, em, s[n];
    const double pi = 3.1415926535897932384626433832795;
  
    d = (double)(n-1)*2;
    for(i=0; i<n; i++) s[i]=cos(pi*i/d);
  
    std::ofstream ficout("out.txt");
  
    // affichage du signal original
    ficout << std::endl  << std::endl << "Signal original :" << std::endl << std::endl;
    for(i=0; i<n; i++) ficout << setprecision(4) << setw(9) << setiosflags(ios_base::fixed) << s[i];
  
    // initialisation
    DaubInit();
  
    // calcul et affichage des coefficients de la transforme
    DaubTrans(s, n);
    ficout << std::endl  << std::endl << "Sa transforme en ondelettes :" << std::endl << std::endl;
    for(i=0; i<n; i++) ficout << setprecision(4) << setw(9) << setiosflags(ios_base::fixed) << s[i];
  
    // calcul et affichage de la transforme inverse
    DaubInvTrans(s, n);
    ficout << std::endl  << std::endl << "Et sa transforme inverse :" << std::endl << std::endl;
    for(i=0; i<n; i++) ficout << setprecision(4) << setw(9) << setiosflags(ios_base::fixed) << s[i];
  
    // calcul et affichage du signal bruit
    br = 0.3;
    s[2] = s[2] + br; 
    ficout << std::endl  << std::endl << "Signal bruit :" << std::endl << std::endl;
    for(i=0; i<n; i++) ficout << setprecision(4) << setw(9) << setiosflags(ios_base::fixed) << s[i];
  
    // calcul et affichage des coefficients de la transforme du signal bruit
    DaubTrans(s, n);
    ficout << std::endl  << std::endl << "Transforme en ondelettes du signal bruit :" << std::endl << std::endl;
    for(i=0; i<n; i++) ficout << setprecision(4) << setw(9) << setiosflags(ios_base::fixed) << s[i];
  
    // on annule les petits coefficients du domaine suprieur
    s[4] = 0.0;
    s[5] = 0.0;
    s[6] = 0.0;
    ficout << std::endl  << std::endl << "Coefficients modifis :" << std::endl << std::endl;
    for(i=0; i<n; i++) ficout << setprecision(4) << setw(9) << setiosflags(ios_base::fixed) << s[i];
  
    // calcul et affichage du signal dbruit
    DaubInvTrans(s, n);
    ficout << std::endl  << std::endl << "Signal dbruit :" << std::endl << std::endl;
    for(i=0; i<n; i++) ficout << setprecision(4) << setw(9) << setiosflags(ios_base::fixed) << s[i];
  
    // affichage de la diffrence entre les 2 signaux
    ficout << std::endl  << std::endl << "cart entre le signal dbruit et l'original :" << std::endl << std::endl;
    for(i=0; i<n; i++) ficout << setprecision(4) << setw(9) << setiosflags(ios_base::fixed) << s[i]-cos(pi*i/d);
  
    // pourcentage entre l'cart maximum et le bruit
    em=0.0;
    for(i=0; i<n; i++) if(em < fabs(s[i]-cos(pi*i/d))) em = fabs(s[i]-cos(pi*i/d));
    ficout << std::endl  << std::endl << "Pourcentage entre l'cart maximum et le bruit : ";
    ficout << setiosflags(ios_base::fixed) << setprecision(1) << 100.0*em/br << " %" << std::endl << std::endl;
    
    ficout.close();
    return 0;
}
