#include "images.h"
//#include <iostream>

void GetPNGtextureInfo(int color_type, gl_texture_t *texinfo) {
     switch (color_type) {
          case PNG_COLOR_TYPE_GRAY:
               texinfo->format = GL_LUMINANCE;
               texinfo->internalFormat = 1;
               break;

          case PNG_COLOR_TYPE_GRAY_ALPHA:
               texinfo->format = GL_LUMINANCE_ALPHA;
               texinfo->internalFormat = 2;
               break;

          case PNG_COLOR_TYPE_RGB:
               texinfo->format = GL_RGB;
               texinfo->internalFormat = 3;
               break;

          case PNG_COLOR_TYPE_RGB_ALPHA:
               texinfo->format = GL_RGBA;
               texinfo->internalFormat = 4;
               break;

          default:
               break;    /* Badness */
     }
}

void GetTGATextureInfo(tga_header_t *header, gl_texture_t *texinfo) {

  texinfo->width = header->width;
  texinfo->height = header->height;

  switch (header->image_type)
    {
    case 3:  /* grayscale 8 bits */
    case 11: /* grayscale 8 bits (RLE) */
      {
	if (header->pixel_depth == 8)
	  {
	    texinfo->format = GL_LUMINANCE;
	    texinfo->internalFormat = 1;
	  }
	else /* 16 bits */
	  {
	    texinfo->format = GL_LUMINANCE_ALPHA;
	    texinfo->internalFormat = 2;
	  }

	break;
      }

    case 1:  /* 8 bits color index */
    case 2:  /* BGR 16-24-32 bits */
    case 9:  /* 8 bits color index (RLE) */
    case 10: /* BGR 16-24-32 bits (RLE) */
      {
	/* 8 bits and 16 bits images will be converted to 24 bits */
	if (header->pixel_depth <= 24)
	  {
	    texinfo->format = GL_RGB;
	    texinfo->internalFormat = 3;
	  }
	else /* 32 bits */
	  {
	    texinfo->format = GL_RGBA;
	    texinfo->internalFormat = 4;
	  }

	break;
      }
    }
}

void ReadTGA8bits(FILE *fp, GLubyte *colormap, gl_texture_t *texinfo)
{
  int i;
  GLubyte color;

  for (i = 0; i < texinfo->width * texinfo->height; ++i)
    {
      /* read index color byte */
      color = (GLubyte)fgetc(fp);

      /* convert to RGB 24 bits */
      texinfo->texels[(i * 3) + 2] = colormap[(color * 3) + 0];
      texinfo->texels[(i * 3) + 1] = colormap[(color * 3) + 1];
      texinfo->texels[(i * 3) + 0] = colormap[(color * 3) + 2];
    }
}

void ReadTGA16bits(FILE *fp, gl_texture_t *texinfo)
{
  int i;
  unsigned short color;

  for (i = 0; i < texinfo->width * texinfo->height; ++i)
    {
      /* read color word */
      color = fgetc(fp) + (fgetc(fp) << 8);

      /* convert BGR to RGB */
      texinfo->texels[(i * 3) + 0] = (GLubyte)(((color & 0x7C00) >> 10) << 3);
      texinfo->texels[(i * 3) + 1] = (GLubyte)(((color & 0x03E0) >>  5) << 3);
      texinfo->texels[(i * 3) + 2] = (GLubyte)(((color & 0x001F) >>  0) << 3);
    }
}

void ReadTGA24bits(FILE *fp, gl_texture_t *texinfo)
{
  int i;

  for (i = 0; i < texinfo->width * texinfo->height; ++i)
    {
      /* read and convert BGR to RGB */
      texinfo->texels[(i * 3) + 2] = (GLubyte)fgetc(fp);
      texinfo->texels[(i * 3) + 1] = (GLubyte)fgetc(fp);
      texinfo->texels[(i * 3) + 0] = (GLubyte)fgetc(fp);
    }
}

void ReadTGA32bits(FILE *fp, gl_texture_t *texinfo)
{
  int i;

  for (i = 0; i < texinfo->width * texinfo->height; ++i)
    {
      /* read and convert BGRA to RGBA */
      texinfo->texels[(i * 4) + 2] = (GLubyte)fgetc(fp);
      texinfo->texels[(i * 4) + 1] = (GLubyte)fgetc(fp);
      texinfo->texels[(i * 4) + 0] = (GLubyte)fgetc(fp);
      texinfo->texels[(i * 4) + 3] = (GLubyte)fgetc(fp);
    }
}

void ReadTGAgray8bits(FILE *fp, gl_texture_t *texinfo)
{
  int i;

  for (i = 0; i < texinfo->width * texinfo->height; ++i)
    {
      /* read grayscale color byte */
      texinfo->texels[i] = (GLubyte)fgetc(fp);
    }
}

void ReadTGAgray16bits(FILE *fp, gl_texture_t *texinfo)
{
  int i;

  for (i = 0; i < texinfo->width * texinfo->height; ++i)
    {
      /* read grayscale color + alpha channel bytes */
      texinfo->texels[(i * 2) + 0] = (GLubyte)fgetc(fp);
      texinfo->texels[(i * 2) + 1] = (GLubyte)fgetc(fp);
    }
}

void ReadTGA8bitsRLE(FILE *fp, GLubyte *colormap, gl_texture_t *texinfo)
{
  int i, size;
  GLubyte color;
  GLubyte packet_header;
  GLubyte *ptr = texinfo->texels;

  while (ptr < texinfo->texels + (texinfo->width * texinfo->height) * 3)
    {
      /* read first byte */
      packet_header = (GLubyte)fgetc(fp);
      size = 1 + (packet_header & 0x7f);

      if (packet_header & 0x80)
	{
	  /* run-length packet */
	  color = (GLubyte)fgetc(fp);

	  for (i = 0; i < size; ++i, ptr += 3)
	    {
	      ptr[0] = colormap[(color * 3) + 2];
	      ptr[1] = colormap[(color * 3) + 1];
	      ptr[2] = colormap[(color * 3) + 0];
	    }
	}
      else
	{
	  /* non run-length packet */
	  for (i = 0; i < size; ++i, ptr += 3)
	    {
	      color = (GLubyte)fgetc(fp);

	      ptr[0] = colormap[(color * 3) + 2];
	      ptr[1] = colormap[(color * 3) + 1];
	      ptr[2] = colormap[(color * 3) + 0];
	    }
	}
   }
}

void ReadTGA16bitsRLE(FILE *fp, gl_texture_t *texinfo)
{
  int i, size;
  unsigned short color;
  GLubyte packet_header;
  GLubyte *ptr = texinfo->texels;

  while (ptr < texinfo->texels + (texinfo->width * texinfo->height) * 3)
    {
      /* read first byte */
      packet_header = fgetc(fp);
      size = 1 + (packet_header & 0x7f);

      if (packet_header & 0x80)
	{
	  /* run-length packet */
	  color = fgetc(fp) + (fgetc(fp) << 8);

	  for (i = 0; i < size; ++i, ptr += 3)
	    {
	      ptr[0] = (GLubyte)(((color & 0x7C00) >> 10) << 3);
	      ptr[1] = (GLubyte)(((color & 0x03E0) >>  5) << 3);
	      ptr[2] = (GLubyte)(((color & 0x001F) >>  0) << 3);
	    }
	}
      else
	{
	  /* non run-length packet */
	  for (i = 0; i < size; ++i, ptr += 3)
	    {
	      color = fgetc(fp) + (fgetc(fp) << 8);

	      ptr[0] = (GLubyte)(((color & 0x7C00) >> 10) << 3);
	      ptr[1] = (GLubyte)(((color & 0x03E0) >>  5) << 3);
	      ptr[2] = (GLubyte)(((color & 0x001F) >>  0) << 3);
	    }
	}
    }
}

void ReadTGA24bitsRLE(FILE *fp, gl_texture_t *texinfo)
{
  int i, size;
  GLubyte rgb[3];
  GLubyte packet_header;
  GLubyte *ptr = texinfo->texels;

  while (ptr < texinfo->texels + (texinfo->width * texinfo->height) * 3)
    {
      /* read first byte */
      packet_header = (GLubyte)fgetc(fp);
      size = 1 + (packet_header & 0x7f);

      if (packet_header & 0x80)
	{
	  /* run-length packet */
	  fread (rgb, sizeof (GLubyte), 3, fp);

	  for (i = 0; i < size; ++i, ptr += 3)
	    {
	      ptr[0] = rgb[2];
	      ptr[1] = rgb[1];
	      ptr[2] = rgb[0];
	    }
	}
      else
	{
	  /* non run-length packet */
	  for (i = 0; i < size; ++i, ptr += 3)
	    {
	      ptr[2] = (GLubyte)fgetc(fp);
	      ptr[1] = (GLubyte)fgetc(fp);
	      ptr[0] = (GLubyte)fgetc(fp);
	    }
	}
    }
}

void ReadTGA32bitsRLE(FILE *fp, gl_texture_t *texinfo)
{
  int i, size;
  GLubyte rgba[4];
  GLubyte packet_header;
  GLubyte *ptr = texinfo->texels;

  while (ptr < texinfo->texels + (texinfo->width * texinfo->height) * 4)
    {
      /* read first byte */
      packet_header = (GLubyte)fgetc(fp);
      size = 1 + (packet_header & 0x7f);

      if (packet_header & 0x80)
	{
	  /* run-length packet */
	  fread (rgba, sizeof (GLubyte), 4, fp);

	  for (i = 0; i < size; ++i, ptr += 4)
	    {
	      ptr[0] = rgba[2];
	      ptr[1] = rgba[1];
	      ptr[2] = rgba[0];
	      ptr[3] = rgba[3];
	    }
	}
      else
	{
	  /* non run-length packet */
	  for (i = 0; i < size; ++i, ptr += 4)
	    {
	      ptr[2] = (GLubyte)fgetc(fp);
	      ptr[1] = (GLubyte)fgetc(fp);
	      ptr[0] = (GLubyte)fgetc(fp);
	      ptr[3] = (GLubyte)fgetc(fp);
	    }
	}
    }
}

void ReadTGAgray8bitsRLE(FILE *fp, gl_texture_t *texinfo)
{
  int i, size;
  GLubyte color;
  GLubyte packet_header;
  GLubyte *ptr = texinfo->texels;

  while (ptr < texinfo->texels + (texinfo->width * texinfo->height))
    {
      /* read first byte */
      packet_header = (GLubyte)fgetc(fp);
      size = 1 + (packet_header & 0x7f);

      if (packet_header & 0x80)
	{
	  /* run-length packet */
	  color = (GLubyte)fgetc(fp);

	  for (i = 0; i < size; ++i, ptr++)
	    *ptr = color;
	}
      else
	{
	  /* non run-length packet */
	  for (i = 0; i < size; ++i, ptr++)
	    *ptr = (GLubyte)fgetc(fp);
	}
    }
}


void ReadTGAgray16bitsRLE(FILE *fp, gl_texture_t *texinfo)
{
  int i, size;
  GLubyte color, alpha;
  GLubyte packet_header;
  GLubyte *ptr = texinfo->texels;

  while (ptr < texinfo->texels + (texinfo->width * texinfo->height) * 2)
    {
      /* read first byte */
      packet_header = (GLubyte)fgetc(fp);
      size = 1 + (packet_header & 0x7f);

      if (packet_header & 0x80)
	{
	  /* run-length packet */
	  color = (GLubyte)fgetc(fp);
	  alpha = (GLubyte)fgetc(fp);

	  for (i = 0; i < size; ++i, ptr += 2)
	    {
	      ptr[0] = color;
	      ptr[1] = alpha;
	    }
	}
      else
	{
	  /* non run-length packet */
	  for (i = 0; i < size; ++i, ptr += 2)
	    {
	      ptr[0] = (GLubyte)fgetc(fp);
	      ptr[1] = (GLubyte)fgetc(fp);
	    }
	}
    }
}

void readGIF(char *szFileName, UCHAR **ucpBits, int *iWidth, int *iHeight) {
FILE       *fi;                   
OLECHAR    wszPath[MAX_PATH+1];   // Pour lire les *.gif voir : Lesson 41 de NeHe  :
IPicture   *pPicture;             // http://nehe.gamedev.net/data/lessons/lesson.asp?lesson=41              
HDC        hdcTemp;               // Cette fonction permet de traiter aussi bien 
HBITMAP    hbmpTemp;              // les types : BMP, EMF, GIF, ICO, JPG et WMF
HRESULT    hr;                    // Donc, readBMP et readJPG sont devenus inutiles.
BITMAPINFO bi = {0};
DWORD      *pBits = 0;
long       i, j, k, n, lWidth, lHeight;
long       lWidthPixels, lHeightPixels;  
unsigned   char *c;	
     fi = fopen(szFileName, "rb");
     if ( fi == NULL ) {rer(21); return;}
     MultiByteToWideChar(CP_ACP, 0, szFileName, -1, wszPath, MAX_PATH);	 // Convert From ASCII To Unicode
     hr = OleLoadPicturePath(wszPath, 0, 0, 0, IID_IPicture, (void**)&pPicture);
     if(FAILED(hr)) {rer(22); return;}
     pPicture->get_Width(&lWidth);	
     pPicture->get_Height(&lHeight);
     hdcTemp = CreateCompatibleDC(GetDC(0));
     if(!hdcTemp) {rer(23); pPicture->Release(); return;}
     lWidthPixels = MulDiv(lWidth, GetDeviceCaps(hdcTemp, LOGPIXELSX), 2540);
     lHeightPixels = MulDiv(lHeight, GetDeviceCaps(hdcTemp, LOGPIXELSY), 2540);
     // Create A Temporary Bitmap
     bi.bmiHeader.biSize          = sizeof(BITMAPINFOHEADER);
     bi.bmiHeader.biBitCount      = 32;
     bi.bmiHeader.biWidth         = lWidthPixels;
     bi.bmiHeader.biHeight        = lHeightPixels;
     bi.bmiHeader.biCompression   = BI_RGB;
     bi.bmiHeader.biPlanes        = 1;
     // Creating A Bitmap This Way Allows Us To Specify Color Depth And Gives Us Imediate Access To The Bits
     hbmpTemp = CreateDIBSection(hdcTemp, &bi, DIB_RGB_COLORS, (void**)&pBits, 0, 0);
     if(!hbmpTemp) {rer(23); pPicture->Release(); DeleteDC(hdcTemp); return;}
     SelectObject(hdcTemp, hbmpTemp);   // Select Handle To Our Temp DC And Our Temp Bitmap Object
     // Render The IPicture On To The Bitmap :
     pPicture->Render(hdcTemp, 0, 0, lWidthPixels, lHeightPixels, 0, lHeight, lWidth, -lHeight, 0);
     n = 3 * lWidthPixels * lHeightPixels;
     *ucpBits = (UCHAR *)malloc(n*sizeof(UCHAR));       
     c = *ucpBits;
     k = lWidthPixels*lHeightPixels-1;               // pour un indice de pBits qui va de 0  k
     for(j=0; j<lHeightPixels; j++) { 
          for(i=0 ; i<lWidthPixels ; i++) {
               BYTE* pPixel = (BYTE*)(&pBits[k-(j*lWidthPixels+i)]);   // on commence par la fin
               n = 3*(j*lWidthPixels+lWidthPixels-1-i);               //  on renverse les lignes    
               *(c+n+0) = pPixel[2];
               *(c+n+1) = pPixel[1];   //  et : BGR -> RGB
               *(c+n+2) = pPixel[0];
          }
     }
     DeleteObject(hbmpTemp);  // donc, ceci efface en mme temps les (BYTE*)&pBits[i]
     DeleteDC(hdcTemp);
     pPicture->Release();
     fclose(fi);
     *iWidth = lWidthPixels;
     *iHeight = lHeightPixels;
     return;
}

void readPNG(char *szFileName, UCHAR **ucpBits, int *iWidth, int *iHeight) {
gl_texture_t *texinfo;
png_byte magic[8];     // http://tfcduke.developpez.com/tutoriel/format/png/
png_structp png_ptr;
png_infop info_ptr;
png_bytep *row_pointers = NULL;
int bit_depth, color_type, i, j, k, n, w, h;
unsigned char *c, *cp;	
FILE *fp = NULL;
     fp = fopen(szFileName, "rb");
     if (fp == NULL) {rer(41); return;}
     fread(magic, 1, sizeof(magic), fp);
     if (png_sig_cmp(magic, 0, sizeof(magic))) {fclose(fp); rer(42); return;}
     png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
     if (!png_ptr) {fclose(fp); rer(43); return;}
     info_ptr = png_create_info_struct(png_ptr);
     if (!info_ptr) {
          png_destroy_read_struct(&png_ptr, NULL, NULL);
          fclose(fp); rer(44); return; }
     texinfo = (gl_texture_t *)malloc(sizeof(gl_texture_t));
     if (setjmp(png_jmpbuf(png_ptr))) {
          png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
          if (row_pointers) free(row_pointers);
          if (texinfo) {
	         if (texinfo->texels) free(texinfo->texels);
               free(texinfo); }
          fclose(fp); rer(45); return; }
     png_init_io(png_ptr, fp);
     png_set_sig_bytes(png_ptr, sizeof(magic));
     png_read_info(png_ptr, info_ptr);
     bit_depth = png_get_bit_depth(png_ptr, info_ptr);
     color_type = png_get_color_type(png_ptr, info_ptr);
     if (color_type == PNG_COLOR_TYPE_PALETTE)
     png_set_palette_to_rgb(png_ptr);
     if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8) png_set_expand_gray_1_2_4_to_8(png_ptr);
     if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) png_set_tRNS_to_alpha(png_ptr);
     if (bit_depth == 16) png_set_strip_16(png_ptr);
     else if (bit_depth < 8) png_set_packing(png_ptr);
     png_read_update_info(png_ptr, info_ptr);
     png_get_IHDR(png_ptr, info_ptr, (png_uint_32*)(&texinfo->width), (png_uint_32*)(&texinfo->height),
                  &bit_depth, &color_type, NULL, NULL, NULL);
     GetPNGtextureInfo(color_type, texinfo);
     w = texinfo->width;
     h = texinfo->height;
     k = texinfo->internalFormat;
     texinfo->texels = (GLubyte *)malloc(sizeof(GLubyte) * w * h * k);
     row_pointers = (png_bytep *)malloc (sizeof (png_bytep) * h);
     for (i = 0; i < h; ++i) row_pointers[i] = (png_bytep)(texinfo->texels + ((h - (i + 1)) * w * k));
     png_read_image(png_ptr, row_pointers);
     png_read_end(png_ptr, NULL);
     png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
     free(row_pointers);
     n = 3 * w * h;
     *ucpBits = (UCHAR *)malloc(n*sizeof(UCHAR));       
     c = *ucpBits;
     for(j=0 ; j<(int)h; j++)
          for(i=0; i<(int)w; i++) {
               cp = (unsigned char *)&texinfo->texels[k*w*h-1-k*(j*w+i)];
               *(c+3*(j*w+w-1-i)+0) = cp[1-k];
               *(c+3*(j*w+w-1-i)+1) = cp[2-k];
               *(c+3*(j*w+w-1-i)+2) = cp[3-k];
          }
     fclose (fp);
     free(texinfo->texels);
     free(texinfo);
     *iWidth = w;
     *iHeight = h;
     return;
}

void readPPM(char *szFileName, UCHAR **ucpBits, int *iWidth, int *iHeight) {
FILE *fi;          
unsigned int w, h, d;          // ici on lit le fichier *.ppm de l'image 
char *t,*ptr;
long fisize;
char buf[128];
int  n, result, code=0;
     fi = fopen(szFileName, "rb");
     if ( fi == NULL ) {rer(51); return;}
     t = fgets(buf, 100, fi);
     if ( t == NULL ) {rer(51); return;}
     if (strncmp(buf, "P3", 2) == 0) code = 3;
     if (strncmp(buf, "P6", 2) == 0) code = 6;
     if(code == 0) {rer(52); return;}
     if(buf[2] == '\n') {
          t = fgets(buf, 100, fi);  // cas gnral : P6\n  ou : P3\n
          ptr=&buf[0];
     }
     else if(buf[2] == '\r') ptr=&buf[3]; else ptr=&buf[2]; 
     while(strncmp(ptr, "#", 1) == 0 ) {   // si commentaires
          t = fgets(buf, 100, fi);
          ptr=&buf[0];
     }
     while(strncmp(ptr, "\n", 1) == 0 ) {  // si ici : '\n' 
          t = fgets(buf, 100, fi);
          ptr=&buf[0];
     }
     d = 0;
     w = strtol(ptr, &ptr, 10);
     h = strtol(ptr, &ptr, 10);
     d = strtol(ptr, &ptr, 10);                  // lecture de d : nb de niveaux
     if ( d != 255 ) {                
          t = fgets(buf, 100, fi);
          d = strtol(buf, &ptr, 10);   //  si ncessaire, dans la ligne suivante
     }
     if ( d != 255 ) {rer(53); return;}          // on se restreint  d = 255 !
     n=3*w*h;                    //  n : nombre d'octets pour les pixels en RGB
     *ucpBits = (UCHAR *)malloc(n*sizeof(UCHAR));  // mmoire des pixels en RGB
     if(code == 3) {            // magic code = P3
          char * buffer;
          size_t result;
          fseek (fi, 0, SEEK_END);
          fisize = ftell (fi);
          rewind(fi);
          buffer = (char *)malloc(sizeof(char)*fisize);
          result = fread(buffer, 1, fisize, fi);
          if(result != fisize) {rer(55); return;} 
          fclose(fi); 
          ptr = buffer + 2;
          char wstr[8], hstr[8], dstr[8]="255";
          sprintf(wstr, "%d", w);
          ptr = strstr(ptr, wstr) + strlen(wstr);
          sprintf(hstr, "%d", h);
          ptr = strstr(ptr, hstr) + strlen(hstr);
          ptr = strstr(ptr, dstr) + 3;
          // Attention : ceci suppose qu'il n'y a pas 255 dans un commentaire inclus !
          unsigned long r, g, b;
          for(unsigned int i=0; i<h; i++)
          for(unsigned int j=0; j<w; j++) {
               while((ptr[0] == '\n')||(ptr[0] == '\r')||(ptr[0] == ' ')) ptr = ptr + 1;
               r = strtol(ptr, &ptr, 10);
               (*ucpBits)[3*(i*w+j)+0] = (unsigned char)r;
               while((ptr[0] == '\n')||(ptr[0] == '\r')||(ptr[0] == ' ')) ptr = ptr + 1;
               g = strtol(ptr, &ptr, 10);
               (*ucpBits)[3*(i*w+j)+1] = (unsigned char)g;
               while((ptr[0] == '\n')||(ptr[0] == '\r')||(ptr[0] == ' ')) ptr = ptr + 1;
               b = strtol(ptr, &ptr, 10);
               (*ucpBits)[3*(i*w+j)+2] = (unsigned char)b;
          }
          free(buffer);
     }
     else {
          fseek(fi, 0, SEEK_END);              // magic code = P6
          fisize = ftell(fi);                  // nombre total d'octets du fichier 
          fseek(fi, fisize-n, SEEK_SET);       // on va lire les n derniers octets
          result = fread(*ucpBits, sizeof(UCHAR), n, fi);
          if(result == 0) {rer(54); return;} 
          fclose(fi); 
     }
     *iWidth=(int)w;   // largeur et
     *iHeight=(int)h;  // hauteur de l'image
     return;
}

void readTGA(char *szFileName, UCHAR **ucpBits, int *iWidth, int *iHeight) {   
gl_texture_t *texinfo;
tga_header_t header;   // http://tfcduke.developpez.com/tutoriel/format/tga/
GLubyte *colormap = NULL;
int i, j, k, n, w, h;
unsigned char *c, *cp;	
FILE *fp;
     fp = fopen(szFileName, "rb");
     if (fp == NULL) {rer(61); return;}
     fread(&header, sizeof(tga_header_t), 1, fp);
     texinfo = (gl_texture_t *)malloc(sizeof(gl_texture_t));
     GetTGATextureInfo(&header, texinfo);
     fseek(fp, header.id_lenght, SEEK_CUR);
     w = texinfo->width;
     h = texinfo->height;
     k = texinfo->internalFormat;
     texinfo->texels = (GLubyte *)malloc(sizeof(GLubyte) * w * h * k);
     if (!texinfo) {
          if (!texinfo->texels) free(texinfo->texels);
          free(texinfo); fclose(fp); rer(62); return; 
     } 
     if (header.colormap_type) {
          colormap = (GLubyte *)malloc (sizeof(GLubyte) * header.cm_length * (header.cm_size >> 3));
          fread(colormap, sizeof(GLubyte), header.cm_length * (header.cm_size >> 3), fp);
     }
     switch (header.image_type) {
          case 0:
               break;
          case 1:
               ReadTGA8bits(fp, colormap, texinfo);
               break;
          case 2:
               switch (header.pixel_depth) {
	              case 16:
	                   ReadTGA16bits(fp, texinfo);
	                   break;
	              case 24:
	                   ReadTGA24bits(fp, texinfo);
	                   break;
	              case 32:
	                   ReadTGA32bits(fp, texinfo);
	                   break;
	         }
               break;
          case 3:
               if (header.pixel_depth == 8) ReadTGAgray8bits(fp, texinfo);
               else ReadTGAgray16bits(fp, texinfo);
               break;
          case 9:
               ReadTGA8bitsRLE(fp, colormap, texinfo);
               break;
          case 10:
               switch (header.pixel_depth) {
	              case 16:
	                   ReadTGA16bitsRLE(fp, texinfo);
	                   break;
	              case 24:
	                   ReadTGA24bitsRLE(fp, texinfo);
	                   break;
	              case 32:
	                   ReadTGA32bitsRLE(fp, texinfo);
	                   break;
	         }
               break;
          case 11:
               if (header.pixel_depth == 8) ReadTGAgray8bitsRLE(fp, texinfo);
               else ReadTGAgray16bitsRLE(fp, texinfo);
               break;
          default:
               free(texinfo->texels);
               free(texinfo);
               texinfo = NULL;
               break;
     }
     if (colormap) free(colormap);
     n = 3 * w * h;
     *ucpBits = (UCHAR *)malloc(n*sizeof(UCHAR));       
     c = *ucpBits;
     for(j=0 ; j<(int)h; j++)
          for(i=0; i<(int)w; i++) {
               cp = (unsigned char *)&texinfo->texels[k*w*h-1-k*(j*w+i)];
               *(c+3*(j*w+w-1-i)+0) = cp[1-k];
               *(c+3*(j*w+w-1-i)+1) = cp[2-k];
               *(c+3*(j*w+w-1-i)+2) = cp[3-k];
          }
     fclose (fp);
     free(texinfo->texels);
     free(texinfo);
     *iWidth = w;   
     *iHeight = h;  
     return;
}

void readTIFF(char *szFileName, UCHAR **ucpBits, int *iWidth, int *iHeight) {        
unsigned int w, h;     
int i,j,n;
unsigned char *cp,*c;
uint32 *raster;          // ici on lit le fichier *.tiff ou *.tif de l'image 
char emsg[1024];
TIFFRGBAImage img;
TIFF *tif;
     tif = TIFFOpen(szFileName, "r");
     if ( tif == NULL ) {rer(71); return;}
     if (TIFFRGBAImageBegin(&img, tif, 0,emsg)){
          n = img.width*img.height;
          raster = (uint32 *)_TIFFmalloc(n*sizeof(uint32));
          if (raster != NULL){
               if (TIFFRGBAImageGet(&img, raster, img.width, img.height) == 0) {rer(72); return;}
          }
          TIFFRGBAImageEnd(&img);
     }
     else {rer(73); return;}
     w = img.width;
     h = img.height;
     *ucpBits = (UCHAR *)malloc(3*n*sizeof(UCHAR));   
     c = *ucpBits;
     for(j=0 ; j<(int)h; j++)
          for(i=0; i<(int)w; i++) {
          cp = (unsigned char *)&raster[n-1-(j*w+i)];
          *(c+3*(j*w+w-1-i)+0) = cp[0];
          *(c+3*(j*w+w-1-i)+1) = cp[1];
          *(c+3*(j*w+w-1-i)+2) = cp[2];
     }
     free(raster);
     raster=NULL;
     TIFFClose(tif);
     *iWidth=(int)w;   // largeur et
     *iHeight=(int)h;  // hauteur de l'image
     return;
}
